/*
 * Time handling and initialization details
 * 
 * Copyright (C) 1991, 1992, 1995  Linus Torvalds
 * Copyright (C) 1996, 1997, 1998  Ralf Baechle
 * Copyright (C) 2000-2002 TiVo Inc. All Rights Reserved.
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/param.h>
#include <linux/interrupt.h>
#include <linux/timex.h>
#include <linux/tivo-monotime.h>

#include <asm/mipsregs.h>
#include <asm/bcm7020.h>

extern int tivo_setup_irq(int, struct irqaction*);
extern volatile unsigned long wall_jiffies;
extern rwlock_t xtime_lock;

#define USECS_PER_JIFFY (1000000/HZ)
// This is (time counter ticks << 15) per jiffy
#define COUNT_PER_JIFFY (((unsigned int)(89701 << 15) + HZ/2) / HZ)

static unsigned int next_comp;

// this must always be called with xtime_lock held and IRQs disabled
static inline unsigned long do_gettimeoffset(void)
{
	unsigned int cur_count, last_comp;
	unsigned long usecs;

	cur_count = *BCM_TIMER_COUNT << 15;
	last_comp = next_comp - COUNT_PER_JIFFY;
	// usecs = (cur_count - last_comp) * USEC_PER_JIFFY / COUNT_PER_JIFFY
	//       = (cur_count - last_comp) * 1000000 / (89701 << 15)
	//       = (cur_count - last_comp) * ((1000000 << 17) / 89701) >> 32
	//       = ((cur_count - last_comp) * 1461210) >> 32
	__asm__(" multu %1, %2;"
	        " mfhi  %0;"
	        : "=r" (usecs)
	        : "r" (cur_count - last_comp), "r" (1461210)
	        : "lo", "hi"
	);
	return usecs;
}

/*
 * This version of gettimeofday has about 11 microsecond precision.
 */
void do_gettimeofday(struct timeval *tv)
{
	unsigned long flags;
	unsigned long usec, sec, lost;

	read_lock_irqsave(&xtime_lock, flags);
	sec = xtime.tv_sec;
	usec = xtime.tv_usec + do_gettimeoffset();
	/*
	 * xtime is atomically updated in timer_bh. jiffies > wall_jiffies
	 * if the timer bottom half hasnt executed yet.
	 */
	lost = jiffies - wall_jiffies;
	read_unlock_irqrestore(&xtime_lock, flags);

	if (lost)
		usec += lost * USECS_PER_JIFFY;

	while (usec >= 1000000) {
		usec -= 1000000;
		sec++;
	}

	tv->tv_sec = sec;
	tv->tv_usec = usec;
}

void do_settimeofday(struct timeval *tv)
{
	write_lock_irq(&xtime_lock);
	/* This is revolting. We need to set the xtime.tv_usec
	 * correctly. However, the value in this location is
	 * is value at the last tick.
	 * Discover what correction gettimeofday
	 * would have done, and then undo it!
	 */
	tv->tv_usec -= do_gettimeoffset();

	if (tv->tv_usec < 0) {
		tv->tv_usec += 1000000;
		tv->tv_sec--;
	}

	xtime = *tv;
	time_adjust = 0;		/* stop active adjtime() */
	time_status |= STA_UNSYNC;
	time_maxerror = NTP_PHASE_LIMIT;
	time_esterror = NTP_PHASE_LIMIT;
	write_unlock_irq(&xtime_lock);
}

/*
 * timer_interrupt() needs to keep up the real-time clock,
 * as well as call the "do_timer()" routine every clocktick
 */
static void timer_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	int cur_count;

	// call do_timer() once, plus once per lost tick
	do {
		do_timer(regs);
		next_comp += COUNT_PER_JIFFY;
		*BCM_TIMER_CMP = next_comp >> 15;
		barrier();
		cur_count = *BCM_TIMER_COUNT << 15;
		barrier();
	} while ((int)(next_comp - cur_count) < 0 &&
	                                     !(*BCM_INTR_STATUS & (1 << 14)));

#if 0 // XXX: RTC write not implemented
	/*
	 * If we have an externally synchronized Linux clock, then update
	 * CMOS clock accordingly every ~11 minutes. Set_rtc_mmss() has to be
	 * called as close as possible to 500 ms before the new second starts.
	 */
	if ((time_status & STA_UNSYNC) == 0 &&
	    xtime.tv_sec > last_rtc_update + 660 &&
	    xtime.tv_usec >= 500000 - ((unsigned) tick) / 2 &&
	    xtime.tv_usec <= 500000 + ((unsigned) tick) / 2)
		if (set_rtc_mmss(xtime.tv_sec) == 0)
			last_rtc_update = xtime.tv_sec;
		else
			last_rtc_update = xtime.tv_sec - 600; /* do it again in 60 s */
#endif
}

/*
 * tivo_time_getcount implements a guaranteed monotonic time counter,
 * using the MIPS cycle counter and an interrupt-driven wrap counter
 */
static volatile unsigned int tivo_time_hi;

void tivo_time_getcount(struct pt_regs *regs)
{
	unsigned int hi1, hi2, lo;

	do {
		hi1 = tivo_time_hi;
		lo = read_32bit_cp0_register(CP0_COUNT);
		hi2 = tivo_time_hi;
	} while (hi1 != hi2);
	regs->regs[2] = hi1;		// high order bits -> v0
	regs->regs[3] = lo;		// low order bits  -> v1
}

/*
 * same monotonic count, for kernel functions
 */
void tivo_monotonic_get_current(monotonic_t *t)
{
	u32 hi1, hi2, lo;

	do {
		hi1 = tivo_time_hi;
		lo = read_32bit_cp0_register(CP0_COUNT);
		hi2 = tivo_time_hi;
	} while (hi1 != hi2);
	t->count = ((u64)hi1 << 32) | lo;
}
EXPORT_SYMBOL(tivo_monotonic_get_current);

static void tivo_time_interrupt(int irq, void *dev_id, struct pt_regs *regs)
{
	tivo_time_hi++;
	write_32bit_cp0_register(CP0_COMPARE,0);
}

static struct irqaction timerint = { timer_interrupt, SA_INTERRUPT, 0, "timer", NULL, NULL };
static struct irqaction tivotime = { tivo_time_interrupt, SA_INTERRUPT, 0, "counter", NULL, NULL };

unsigned long tivo_counts_per_256usec;
EXPORT_SYMBOL(tivo_counts_per_256usec);

void __init time_init(void)
{
	unsigned long count1, count2, old_tcpu1, old_tcpu2;
	unsigned long flags;
	int bail_count;

	// XXX: this should init from RTC instead, but for now, just
	// init to Jan 02-Jan-1970 00:00, so don't get negative time,
	// even after adjusting to local time
	xtime.tv_sec = 24*3600;
	xtime.tv_usec = 0;

	// calibrate the monotonic time counter
	old_tcpu1 = old_tcpu2 = 0;
	bail_count = 500;
	save_and_cli(flags);
	*BCM_TIMER_COUNT = 0;
	do {
		u32 comp;
		int i;

		old_tcpu2= old_tcpu1;
		old_tcpu1 = tivo_counts_per_256usec;
		comp = *BCM_TIMER_COUNT;
		while (*BCM_TIMER_COUNT == comp) barrier();
		count1 = read_32bit_cp0_register(CP0_COUNT);
		// this is about 350 microseconds, so we should get a count
		// that is accurate to +/- 1 cycle per 256 microseconds
		for (i = 0; i < 32; i++) {
		    comp = *BCM_TIMER_COUNT;
		    while (*BCM_TIMER_COUNT == comp) barrier();
		}
		count2 = read_32bit_cp0_register(CP0_COUNT);
		// the math here has enough precision to handle 27869 counts
		// per microsecond without overflow, which is plenty for now,
		// at least until we get that 56GHz CPU...
		tivo_counts_per_256usec = ((count2 - count1) * 432 + 301) / 602;
	} while ((tivo_counts_per_256usec != old_tcpu1) || 
		 (old_tcpu1 != old_tcpu2) ||
		 (--bail_count == 0));
	restore_flags(flags);
	printk("Monotonic time calibrated: %ld.%02ld counts per usec\n",
	       tivo_counts_per_256usec / 256,
	       ((tivo_counts_per_256usec & 0xff) * 100) / 256);
	
	// set for .01s = 897/89.7KHz cycle
	next_comp = COUNT_PER_JIFFY;
	*BCM_TIMER_COUNT = 0;
	*BCM_TIMER_CMP = next_comp >> 15;
	tivo_setup_irq(BCM_IRQ_TMR_COMP, &timerint);

	write_32bit_cp0_register(CP0_COMPARE,0);
	tivo_setup_irq(CPU_IRQ_INT5, &tivotime);
}

