/*
 *  linux/drivers/char/mem.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  Added devfs support. 
 *    Jan-11-1998, C. Scott Ananian <cananian@alumni.princeton.edu>
 *  Shared /dev/zero mmaping support, Feb 2000, Kanoj Sarcar <kanoj@sgi.com>
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/mm.h>
#include <linux/miscdevice.h>
#include <linux/tpqic02.h>
#include <linux/ftape.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/mman.h>
#include <linux/random.h>
#include <linux/init.h>
#include <linux/raw.h>
#include <linux/tty.h>
#include <linux/capability.h>

#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/pgalloc.h>

#ifdef CONFIG_I2C
extern int i2c_init_all(void);
#endif
#ifdef CONFIG_FB
extern void fbmem_init(void);
#endif
#ifdef CONFIG_PROM_CONSOLE
extern void prom_con_init(void);
#endif
#ifdef CONFIG_MDA_CONSOLE
extern void mda_console_init(void);
#endif
#if defined(CONFIG_S390_TAPE) && defined(CONFIG_S390_TAPE_CHAR)
extern void tapechar_init(void);
#endif

#ifdef CONFIG_CONTIGMEM
#include <linux/contigmem.h>
#include <linux/bootmem.h>
#include <linux/tivoconfig.h>

extern void _end;

#ifdef CONFIG_TIVO_DEVEL
// Amount of memory to protect after each (non-zero sized) contigmem region
#define CONTIGMEM_BARRIER_SIZE                  PAGE_SIZE
// 128 KB process log
#define DEFAULT_CONTIGMEM_PLOG_SIZE             (128 * 1024)
#else 
// For release, no plog and no contigmem barriers
#define CONTIGMEM_BARRIER_SIZE                 ( 0 )
#define DEFAULT_CONTIGMEM_PLOG_SIZE            ( 0 )
#endif

// Guard signature to fill contigmem barrier with
#define CONTIGMEM_BARRIER_BYTE                  (0xCB)

#define CONTIGMEM_MINLEFT			(4 * 1024 * 1024)

unsigned long contigmem_start = 0;
unsigned long contigmem_size[MAX_CONTIGMEM_REGION + 1];
unsigned long contigmem_barrier = CONTIGMEM_BARRIER_SIZE;

static int __init contigmem_setup(char *str)
{
	unsigned long override_size;
        int region = -1;
	char *dummy;

        if (!str || *str == '\0') { 
            return 0;
        }

        // Supports original format (contigmem=YYY) to override
        // total amount of contigmem. Excess is added to last region
        if (*str == '=') {
	    str++;
	    override_size = (unsigned long)memparse(str, &dummy);
	    SetTivoConfig(kTivoConfigContigmemSize, override_size);
        }
        // Alternative format where you can override individual regions
        // using "contigmemX=YYY". Regions are numbered from 0
        else {
            enum TivoConfigSelectors tcs;
            
            if (sscanf(str, "%d=", &region) != 1) {
                return 0;
            }
            if (region < 0 || region > MAX_CONTIGMEM_REGION) {
                return 0;
            }
            while (*str != '=') {
                str++;
            }
            str++;

            override_size = (unsigned long)memparse(str, &dummy);
            
            // don't look!
            tcs = (enum TivoConfigSelectors)((unsigned int)kTivoConfigContigmemRegion0 + region);
            SetTivoConfig(tcs, override_size);
        }
	return 1;
}

static int __init contigbarrier_setup(char *str)
{
        unsigned long override_size;
        char *dummy;
        
        override_size = (unsigned long)memparse(str, &dummy);
        contigmem_barrier = PAGE_ALIGN(override_size);
        return 1;
}

__setup("contigmem", contigmem_setup);
__setup("contigbarrier=", contigbarrier_setup);

void __init contigmem(void) {
	unsigned long physmem_top = max_low_pfn * PAGE_SIZE;
	unsigned long contigmem_top;
        unsigned long cm_offset;
	unsigned long contigmem_total = 0;
	TivoConfigValue contigmem_set;
	int i, cms_top = 0;

	/* Cap to 32MB for this calculation */
	if (physmem_top > 32 * 1024 * 1024) {
		contigmem_top = (unsigned long)__va(32 * 1024 * 1024);
	} else {
		contigmem_top = (unsigned long)__va(physmem_top);
	}
        
        contigmem_size[CONTIGMEM_REGION_PLOG] = DEFAULT_CONTIGMEM_PLOG_SIZE;

	/*
	 * If the query fails on the plog region, just use the
	 * hardcoded default. This is the normal case.
	 */
	for (i = 0; i <= MAX_CONTIGMEM_REGION; i++) {
		if (!GetTivoConfig(kTivoConfigContigmemRegion0 + i,
		                   &contigmem_size[i])) {
			cms_top = i;
		}
		contigmem_total += contigmem_size[i];
		if (contigmem_size[i]) {
                        // place barrier after a nonzero-sized region
                        contigmem_total += contigmem_barrier;
		}
	}
	/* Add any excess to the last defined region */
	if (!GetTivoConfig(kTivoConfigContigmemSize, &contigmem_set) &&
	    contigmem_set != contigmem_total &&
	    contigmem_set + contigmem_size[cms_top] >= contigmem_total) {
		contigmem_size[cms_top] += contigmem_set - contigmem_total;
		printk("Adjusted region %d to %ld bytes\n",
		       	cms_top, contigmem_size[cms_top]);
		contigmem_total = contigmem_set;
	}

	contigmem_start = PAGE_ALIGN(contigmem_top - contigmem_total);
	if (contigmem_total < contigmem_top - contigmem_start) {
		contigmem_size[cms_top] += contigmem_top - contigmem_start -
		                           contigmem_total;
		contigmem_total = contigmem_top - contigmem_start;
	}

	/* This is an approximate check at best... */
	if (contigmem_start < (unsigned long)&_end + CONTIGMEM_MINLEFT) {
		contigmem_start = PAGE_ALIGN((unsigned long)&_end + CONTIGMEM_MINLEFT);
		if (contigmem_start + contigmem_total > (unsigned long)__va(physmem_top)) {
			contigmem_start = 0;
			for (i = 0; i <= MAX_CONTIGMEM_REGION; i++) {
				contigmem_size[i] = 0;
			}
			printk("Insufficient real memory to allocate contiguous region.\n");
			return;
		}
	}
	/* In case it wasn't set, update the CMSZ config entry */
	SetTivoConfig(kTivoConfigContigmemSize, contigmem_total);
	reserve_bootmem(__pa(contigmem_start), contigmem_total);

        /* Print offsets and initialize barrier signature */
        cm_offset = contigmem_start;
        for (i = 0; i <= cms_top; ++i) { 
            if (contigmem_size[i]) {
                printk("Contiguous region %d: %ld bytes @ address 0x%lx\n",
                       i, contigmem_size[i], cm_offset);
                cm_offset += contigmem_size[i];
                if (contigmem_barrier) {
                    memset((void *)cm_offset, CONTIGMEM_BARRIER_BYTE, contigmem_barrier);
                    printk("Contigmem barrier %ld bytes at 0x%lx.\n",
                           contigmem_barrier, cm_offset);
                    cm_offset += contigmem_barrier;
                }
            }
        }

	printk("Contiguous region of %ld bytes total reserved at 0x%lx.\n",
	       contigmem_total, contigmem_start);
}

// This is the exported interface for modules
//
// Regions are numbered beginning from 0
//
int mem_get_contigmem_region(unsigned int region, unsigned long *start,
                             unsigned long *size)
{
	unsigned int i;

	if (region <= MAX_CONTIGMEM_REGION) {
		*start = contigmem_start;
		for (i = 0; i < region; i++) {
			*start += contigmem_size[i];
                        // account for the barrier between each non empty region
                        if (contigmem_size[i]) {
                            *start += contigmem_barrier;
                        }
		}
		*size = contigmem_size[region];

		// this is compatibility cruft, redirect region 3 to 8 if
		// 3 is empty
		if ( region == 3 && contigmem_size[region] == 0 ) {
			return mem_get_contigmem_region( 8, start, size );
		}

		return 0;
	} else {
		return -EINVAL;
	}
}
EXPORT_SYMBOL(mem_get_contigmem_region);

static int mem_ioctl(struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg)
{
	unsigned int region;
        unsigned long region_start = 0;
        unsigned long region_size = 0;
	struct contigmem_region_request *request =
	    (struct contigmem_region_request *)arg;
	int ret;

	if ( contigmem_start == 0 )
		return -ENOMEM;

	switch (_IOC_NR(cmd)) {
		/* 1,2,4,5 are compatibiliy cruft... */
		case 1:
		case 2:
			region = 1;
			break;
		case 4:
		case 5:
			region = 2;
			break;
		default:
			if ( get_user( region, &request->region ) < 0 )
				return -EFAULT;
			break;
	}

	ret = mem_get_contigmem_region( region, &region_start, &region_size );
	if ( ret < 0 )
		return ret;

	region_start = virt_to_phys( (void *)region_start );

	if ( cmd == CONTIGMEM_GET_REGION ) {
		if ( put_user( region_start, &request->phys_start ) < 0 ||
		     put_user( region_size, &request->size ) < 0 )
			return -EFAULT;
		return 0;
	} else {
		/* even more compatibility cruft */
		switch (_IOC_NR(cmd)) {
			case 1:
			case 4:
				return (int)region_start;
			case 2:
			case 5:
				return (int)region_size;
			default:
				break;
		}
	}

	return -EINVAL;
}
#endif /* CONFIG_CONTIGMEM */

static ssize_t do_write_mem(struct file * file, void *p, unsigned long realp,
			    const char * buf, size_t count, loff_t *ppos)
{
	ssize_t written;

	written = 0;
#if defined(__sparc__) || defined(__mc68000__)
	/* we don't have page 0 mapped on sparc and m68k.. */
	if (realp < PAGE_SIZE) {
		unsigned long sz = PAGE_SIZE-realp;
		if (sz > count) sz = count; 
		/* Hmm. Do something? */
		buf+=sz;
		p+=sz;
		count-=sz;
		written+=sz;
	}
#endif
	if (copy_from_user(p, buf, count))
		return -EFAULT;
	written += count;
	*ppos += written;
	return written;
}


/*
 * This funcion reads the *physical* memory. The f_pos points directly to the 
 * memory location. 
 */
static ssize_t read_mem(struct file * file, char * buf,
			size_t count, loff_t *ppos)
{
	unsigned long p = *ppos;
	unsigned long end_mem;
	ssize_t read;
	
	end_mem = __pa(high_memory);
	if (p >= end_mem)
		return 0;
	if (count > end_mem - p)
		count = end_mem - p;
	read = 0;
#if defined(__sparc__) || defined(__mc68000__)
	/* we don't have page 0 mapped on sparc and m68k.. */
	if (p < PAGE_SIZE) {
		unsigned long sz = PAGE_SIZE-p;
		if (sz > count) 
			sz = count; 
		if (sz > 0) {
			if (clear_user(buf, sz))
				return -EFAULT;
			buf += sz; 
			p += sz; 
			count -= sz; 
			read += sz; 
		}
	}
#endif
	if (copy_to_user(buf, __va(p), count))
		return -EFAULT;
	read += count;
	*ppos += read;
	return read;
}

static ssize_t write_mem(struct file * file, const char * buf, 
			 size_t count, loff_t *ppos)
{
	unsigned long p = *ppos;
	unsigned long end_mem;

	end_mem = __pa(high_memory);
	if (p >= end_mem)
		return 0;
	if (count > end_mem - p)
		count = end_mem - p;
	return do_write_mem(file, __va(p), p, buf, count, ppos);
}

#ifndef pgprot_noncached

/*
 * This should probably be per-architecture in <asm/pgtable.h>
 */
static inline pgprot_t pgprot_noncached(pgprot_t _prot)
{
	unsigned long prot = pgprot_val(_prot);

#if defined(__i386__) || defined(__x86_64__)
	/* On PPro and successors, PCD alone doesn't always mean 
	    uncached because of interactions with the MTRRs. PCD | PWT
	    means definitely uncached. */ 
	if (boot_cpu_data.x86 > 3)
		prot |= _PAGE_PCD | _PAGE_PWT;
#elif defined(__powerpc__)
	prot |= _PAGE_NO_CACHE | _PAGE_GUARDED;
#elif defined(__mc68000__)
#ifdef SUN3_PAGE_NOCACHE
	if (MMU_IS_SUN3)
		prot |= SUN3_PAGE_NOCACHE;
	else
#endif
	if (MMU_IS_851 || MMU_IS_030)
		prot |= _PAGE_NOCACHE030;
	/* Use no-cache mode, serialized */
	else if (MMU_IS_040 || MMU_IS_060)
		prot = (prot & _CACHEMASK040) | _PAGE_NOCACHE_S;
#endif

	return __pgprot(prot);
}

#endif /* !pgprot_noncached */

/*
 * Architectures vary in how they handle caching for addresses 
 * outside of main memory.
 */
static inline int noncached_address(unsigned long addr)
{
#if defined(__i386__)
	/* 
	 * On the PPro and successors, the MTRRs are used to set
	 * memory types for physical addresses outside main memory, 
	 * so blindly setting PCD or PWT on those pages is wrong.
	 * For Pentiums and earlier, the surround logic should disable 
	 * caching for the high addresses through the KEN pin, but
	 * we maintain the tradition of paranoia in this code.
	 */
 	return !( test_bit(X86_FEATURE_MTRR, &boot_cpu_data.x86_capability) ||
		  test_bit(X86_FEATURE_K6_MTRR, &boot_cpu_data.x86_capability) ||
		  test_bit(X86_FEATURE_CYRIX_ARR, &boot_cpu_data.x86_capability) ||
		  test_bit(X86_FEATURE_CENTAUR_MCR, &boot_cpu_data.x86_capability) )
	  && addr >= __pa(high_memory);
#else
	return addr >= __pa(high_memory);
#endif
}

static int mmap_mem(struct file * file, struct vm_area_struct * vma)
{
	unsigned long offset = vma->vm_pgoff << PAGE_SHIFT;

	/*
	 * Accessing memory above the top the kernel knows about or
	 * through a file pointer that was marked O_SYNC will be
	 * done non-cached.
	 */
	if (noncached_address(offset) || (file->f_flags & O_SYNC))
		vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);

	/* Don't try to swap out physical pages.. */
	vma->vm_flags |= VM_RESERVED;

	/*
	 * Don't dump addresses that are not real memory to a core file.
	 */
	if (offset >= __pa(high_memory) || (file->f_flags & O_SYNC))
		vma->vm_flags |= VM_IO;

	if (remap_page_range(vma->vm_start, offset, vma->vm_end-vma->vm_start,
			     vma->vm_page_prot))
		return -EAGAIN;
	return 0;
}

/*
 * This function reads the *virtual* memory as seen by the kernel.
 */
static ssize_t read_kmem(struct file *file, char *buf, 
			 size_t count, loff_t *ppos)
{
	unsigned long p = *ppos;
	ssize_t read = 0;
	ssize_t virtr = 0;
	char * kbuf; /* k-addr because vread() takes vmlist_lock rwlock */
		
	if (p < (unsigned long) high_memory) {
		read = count;
		if (count > (unsigned long) high_memory - p)
			read = (unsigned long) high_memory - p;

#if defined(__sparc__) || defined(__mc68000__)
		/* we don't have page 0 mapped on sparc and m68k.. */
		if (p < PAGE_SIZE && read > 0) {
			size_t tmp = PAGE_SIZE - p;
			if (tmp > read) tmp = read;
			if (clear_user(buf, tmp))
				return -EFAULT;
			buf += tmp;
			p += tmp;
			read -= tmp;
			count -= tmp;
		}
#endif
		if (copy_to_user(buf, (char *)p, read))
			return -EFAULT;
		p += read;
		buf += read;
		count -= read;
	}

	if (count > 0) {
		kbuf = (char *)__get_free_page(GFP_KERNEL);
		if (!kbuf)
			return -ENOMEM;
		while (count > 0) {
			int len = count;

			if (len > PAGE_SIZE)
				len = PAGE_SIZE;
			len = vread(kbuf, (char *)p, len);
			if (!len)
				break;
			if (copy_to_user(buf, kbuf, len)) {
				free_page((unsigned long)kbuf);
				return -EFAULT;
			}
			count -= len;
			buf += len;
			virtr += len;
			p += len;
		}
		free_page((unsigned long)kbuf);
	}
 	*ppos = p;
 	return virtr + read;
}

extern long vwrite(char *buf, char *addr, unsigned long count);

/*
 * This function writes to the *virtual* memory as seen by the kernel.
 */
static ssize_t write_kmem(struct file * file, const char * buf, 
			  size_t count, loff_t *ppos)
{
	unsigned long p = *ppos;
	ssize_t wrote = 0;
	ssize_t virtr = 0;
	char * kbuf; /* k-addr because vwrite() takes vmlist_lock rwlock */

	if (p < (unsigned long) high_memory) {
		wrote = count;
		if (count > (unsigned long) high_memory - p)
			wrote = (unsigned long) high_memory - p;

		wrote = do_write_mem(file, (void*)p, p, buf, wrote, ppos);

		p += wrote;
		buf += wrote;
		count -= wrote;
	}

	if (count > 0) {
		kbuf = (char *)__get_free_page(GFP_KERNEL);
		if (!kbuf)
			return -ENOMEM;
		while (count > 0) {
			int len = count;

			if (len > PAGE_SIZE)
				len = PAGE_SIZE;
			if (len && copy_from_user(kbuf, buf, len)) {
				free_page((unsigned long)kbuf);
				return -EFAULT;
			}
			len = vwrite(kbuf, (char *)p, len);
			count -= len;
			buf += len;
			virtr += len;
			p += len;
		}
		free_page((unsigned long)kbuf);
	}

 	*ppos = p;
 	return virtr + wrote;
}

#if defined(CONFIG_ISA) || !defined(__mc68000__)
static ssize_t read_port(struct file * file, char * buf,
			 size_t count, loff_t *ppos)
{
	unsigned long i = *ppos;
	char *tmp = buf;

	if (verify_area(VERIFY_WRITE,buf,count))
		return -EFAULT; 
	while (count-- > 0 && i < 65536) {
		if (__put_user(inb(i),tmp) < 0) 
			return -EFAULT;  
		i++;
		tmp++;
	}
	*ppos = i;
	return tmp-buf;
}

static ssize_t write_port(struct file * file, const char * buf,
			  size_t count, loff_t *ppos)
{
	unsigned long i = *ppos;
	const char * tmp = buf;

	if (verify_area(VERIFY_READ,buf,count))
		return -EFAULT;
	while (count-- > 0 && i < 65536) {
		char c;
		if (__get_user(c, tmp)) 
			return -EFAULT; 
		outb(c,i);
		i++;
		tmp++;
	}
	*ppos = i;
	return tmp-buf;
}
#endif

static ssize_t read_null(struct file * file, char * buf,
			 size_t count, loff_t *ppos)
{
	return 0;
}

static ssize_t write_null(struct file * file, const char * buf,
			  size_t count, loff_t *ppos)
{
	return count;
}

/*
 * For fun, we are using the MMU for this.
 */
static inline size_t read_zero_pagealigned(char * buf, size_t size)
{
	struct mm_struct *mm;
	struct vm_area_struct * vma;
	unsigned long addr=(unsigned long)buf;

	mm = current->mm;
	/* Oops, this was forgotten before. -ben */
	down_read(&mm->mmap_sem);

	/* For private mappings, just map in zero pages. */
	for (vma = find_vma(mm, addr); vma; vma = vma->vm_next) {
		unsigned long count;

		if (vma->vm_start > addr || (vma->vm_flags & VM_WRITE) == 0)
			goto out_up;
		if (vma->vm_flags & VM_SHARED)
			break;
		count = vma->vm_end - addr;
		if (count > size)
			count = size;

		zap_page_range(mm, addr, count);
        	zeromap_page_range(addr, count, PAGE_COPY);

		size -= count;
		buf += count;
		addr += count;
		if (size == 0)
			goto out_up;
	}

	up_read(&mm->mmap_sem);
	
	/* The shared case is hard. Let's do the conventional zeroing. */ 
	do {
		unsigned long unwritten = clear_user(buf, PAGE_SIZE);
		if (unwritten)
			return size + unwritten - PAGE_SIZE;
		if (current->need_resched)
			schedule();
		buf += PAGE_SIZE;
		size -= PAGE_SIZE;
	} while (size);

	return size;
out_up:
	up_read(&mm->mmap_sem);
	return size;
}

static ssize_t read_zero(struct file * file, char * buf, 
			 size_t count, loff_t *ppos)
{
	unsigned long left, unwritten, written = 0;

	if (!count)
		return 0;

	if (!access_ok(VERIFY_WRITE, buf, count))
		return -EFAULT;

	left = count;

	/* do we want to be clever? Arbitrary cut-off */
	if (count >= PAGE_SIZE*4) {
		unsigned long partial;

		/* How much left of the page? */
		partial = (PAGE_SIZE-1) & -(unsigned long) buf;
		unwritten = clear_user(buf, partial);
		written = partial - unwritten;
		if (unwritten)
			goto out;
		left -= partial;
		buf += partial;
		unwritten = read_zero_pagealigned(buf, left & PAGE_MASK);
		written += (left & PAGE_MASK) - unwritten;
		if (unwritten)
			goto out;
		buf += left & PAGE_MASK;
		left &= ~PAGE_MASK;
	}
	unwritten = clear_user(buf, left);
	written += left - unwritten;
out:
	return written ? written : -EFAULT;
}

static int mmap_zero(struct file * file, struct vm_area_struct * vma)
{
	if (vma->vm_flags & VM_SHARED)
		return shmem_zero_setup(vma);
	if (zeromap_page_range(vma->vm_start, vma->vm_end - vma->vm_start, vma->vm_page_prot))
		return -EAGAIN;
	return 0;
}

static ssize_t write_full(struct file * file, const char * buf,
			  size_t count, loff_t *ppos)
{
	return -ENOSPC;
}

/*
 * Special lseek() function for /dev/null and /dev/zero.  Most notably, you
 * can fopen() both devices with "a" now.  This was previously impossible.
 * -- SRB.
 */

static loff_t null_lseek(struct file * file, loff_t offset, int orig)
{
	return file->f_pos = 0;
}

/*
 * The memory devices use the full 32/64 bits of the offset, and so we cannot
 * check against negative addresses: they are ok. The return value is weird,
 * though, in that case (0).
 *
 * also note that seeking relative to the "end of file" isn't supported:
 * it has no meaning, so it returns -EINVAL.
 */
static loff_t memory_lseek(struct file * file, loff_t offset, int orig)
{
	switch (orig) {
		case 0:
			file->f_pos = offset;
			return file->f_pos;
		case 1:
			file->f_pos += offset;
			return file->f_pos;
		default:
			return -EINVAL;
	}
}

static int open_port(struct inode * inode, struct file * filp)
{
	return capable(CAP_SYS_RAWIO) ? 0 : -EPERM;
}

#define mmap_kmem	mmap_mem
#define zero_lseek	null_lseek
#define full_lseek      null_lseek
#define write_zero	write_null
#define read_full       read_zero
#define open_mem	open_port
#define open_kmem	open_mem

static struct file_operations mem_fops = {
	llseek:		memory_lseek,
	read:		read_mem,
	write:		write_mem,
#ifdef CONFIG_CONTIGMEM
	ioctl:		mem_ioctl,
#endif
	mmap:		mmap_mem,
	open:		open_mem,
};

static struct file_operations kmem_fops = {
	llseek:		memory_lseek,
	read:		read_kmem,
	write:		write_kmem,
	mmap:		mmap_kmem,
	open:		open_kmem,
};

static struct file_operations null_fops = {
	llseek:		null_lseek,
	read:		read_null,
	write:		write_null,
};

#if defined(CONFIG_ISA) || !defined(__mc68000__)
static struct file_operations port_fops = {
	llseek:		memory_lseek,
	read:		read_port,
	write:		write_port,
	open:		open_port,
};
#endif

static struct file_operations zero_fops = {
	llseek:		zero_lseek,
	read:		read_zero,
	write:		write_zero,
	mmap:		mmap_zero,
};

static struct file_operations full_fops = {
	llseek:		full_lseek,
	read:		read_full,
	write:		write_full,
};

#ifdef CONFIG_TIVO
#ifdef CONFIG_4xx
#define sercons_poll s403_poll
#endif
static int cons_open(struct inode *inode, struct file *file)
{
    return 0;
}

static ssize_t write_cons(struct file *file, const char *buf,
        size_t count, loff_t *ppos) {
    char c;
    const char *p = buf;
    int i = 0;

    for (; i < count; i++, p++) {
        get_user(c, p);
        printk("%c", c);
    }
    return i;
}

static ssize_t read_cons(struct file *file, char *buf,
        size_t count, loff_t *ppos) {
    char *p = buf;
    int i = 0;
    extern int sercons_poll(void);

    for (; i < count; i++, p++) {
	    int c;
	    while ((c = sercons_poll()) == -1) {
            schedule_timeout(1);
	    }
	    if (put_user(c, p))
		break;
	    if (c == '\n')
		break;
    }
    return i;
}

struct file_operations cons_fops = {
	read:		read_cons,
	write:		write_cons,
	open:		cons_open,
};
#endif

static int memory_open(struct inode * inode, struct file * filp)
{
	switch (MINOR(inode->i_rdev)) {
		case 1:
#ifdef CONFIG_CONTIGMEM
		case 11:
#endif
			filp->f_op = &mem_fops;
			break;
		case 2:
			filp->f_op = &kmem_fops;
			break;
		case 3:
			filp->f_op = &null_fops;
			break;
#if defined(CONFIG_ISA) || !defined(__mc68000__)
		case 4:
			filp->f_op = &port_fops;
			break;
#endif
		case 5:
			filp->f_op = &zero_fops;
			break;
		case 7:
			filp->f_op = &full_fops;
			break;
		case 8:
			filp->f_op = &random_fops;
			break;
		case 9:
			filp->f_op = &urandom_fops;
			break;
#ifdef CONFIG_TIVO
		case 100:
			filp->f_op = &cons_fops;
			break;
#endif
		default:
			return -ENXIO;
	}
	if (filp->f_op && filp->f_op->open)
		return filp->f_op->open(inode,filp);
	return 0;
}

void __init memory_devfs_register (void)
{
    /*  These are never unregistered  */
    static const struct {
	unsigned short minor;
	char *name;
	umode_t mode;
	struct file_operations *fops;
    } list[] = { /* list of minor devices */
	{1, "mem",     S_IRUSR | S_IWUSR | S_IRGRP, &mem_fops},
	{2, "kmem",    S_IRUSR | S_IWUSR | S_IRGRP, &kmem_fops},
	{3, "null",    S_IRUGO | S_IWUGO,           &null_fops},
#if defined(CONFIG_ISA) || !defined(__mc68000__)
	{4, "port",    S_IRUSR | S_IWUSR | S_IRGRP, &port_fops},
#endif
	{5, "zero",    S_IRUGO | S_IWUGO,           &zero_fops},
	{7, "full",    S_IRUGO | S_IWUGO,           &full_fops},
	{8, "random",  S_IRUGO | S_IWUSR,           &random_fops},
	{9, "urandom", S_IRUGO | S_IWUSR,           &urandom_fops}
    };
    int i;

    for (i=0; i<(sizeof(list)/sizeof(*list)); i++)
	devfs_register (NULL, list[i].name, DEVFS_FL_NONE,
			MEM_MAJOR, list[i].minor,
			list[i].mode | S_IFCHR,
			list[i].fops, NULL);
}

static struct file_operations memory_fops = {
	open:		memory_open,	/* just a selector for the real open */
};

int __init chr_dev_init(void)
{
	if (devfs_register_chrdev(MEM_MAJOR,"mem",&memory_fops))
		printk("unable to get major %d for memory devs\n", MEM_MAJOR);
	memory_devfs_register();
	rand_initialize();
#ifdef CONFIG_I2C
	i2c_init_all();
#endif
#if defined (CONFIG_FB)
	fbmem_init();
#endif
#if defined (CONFIG_PROM_CONSOLE)
	prom_con_init();
#endif
#if defined (CONFIG_MDA_CONSOLE)
	mda_console_init();
#endif
	tty_init();
#ifdef CONFIG_M68K_PRINTER
	lp_m68k_init();
#endif
	misc_init();
#if CONFIG_QIC02_TAPE
	qic02_tape_init();
#endif
#ifdef CONFIG_FTAPE
	ftape_init();
#endif
#if defined(CONFIG_S390_TAPE) && defined(CONFIG_S390_TAPE_CHAR)
	tapechar_init();
#endif
	return 0;
}

__initcall(chr_dev_init);
