//depot/mainline/os/linux-2.4/drivers/usb/pegasus.c#13 - integrate change 102434 (text)
/*
 *  Copyright (c) 1999-2002 Petko Manolov (petkan@users.sourceforge.net)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *	
 *
 *	ChangeLog:
 *		....	Most of the time spend reading sources & docs.
 *		v0.2.x	First official release for the Linux kernel.
 *		v0.3.0	Beutified and structured, some bugs fixed.
 *		v0.3.x	URBifying bulk requests and bugfixing. First relatively
 *			stable release. Still can touch device's registers only
 *			from top-halves.
 *		v0.4.0	Control messages remained unurbified are now URBs.
 *			Now we can touch the HW at any time.
 *		v0.4.9	Control urbs again use process context to wait. Argh...
 *			Some long standing bugs (enable_net_traffic) fixed.
 *			Also nasty trick about resubmiting control urb from
 *			interrupt context used. Please let me know how it
 *			behaves. Pegasus II support added since this version.
 *			TODO: suppressing HCD warnings spewage on disconnect.
 *		v0.4.13	Ethernet address is now set at probe(), not at open()
 *			time as this seems to break dhcpd. 
 *		v0.4.25	ethtool support added.
 */

#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/delay.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/ethtool.h>
#include <linux/mii.h>
#include <linux/usb.h>
#include <linux/module.h>
#include <asm/uaccess.h>
#include "pegasus.h"

#ifdef DEBUG
#undef dbg
#define dbg(format, arg...) printk(KERN_INFO __FILE__ "!"  __FUNCTION__ ": " format "\n" , ## arg)
static int msglvl;
#endif

/*
 * Version Information
 */
#define DRIVER_VERSION "v0.4.26 (2002/03/21)"
#define DRIVER_AUTHOR "Petko Manolov <petkan@users.sourceforge.net>"
#define DRIVER_DESC "Pegasus/Pegasus II USB Ethernet driver"

#ifdef PEGASUS_USE_INTR
static void disable_net_traffic_sched( void *pegasus );
static void enable_net_traffic_sched( void *pegasus ); 
#endif

static void disable_net_traffic( pegasus_t *pegasus );
static void enable_net_traffic( pegasus_t *pegasus ); 

static void read_bulk_callback( struct urb *urb );

#undef	PEGASUS_WRITE_EEPROM
#define	BMSR_MEDIA	(BMSR_10HALF | BMSR_10FULL | BMSR_100HALF | \
			BMSR_100FULL | BMSR_ANEGCAPABLE)

static int loopback = 0;
static int mii_mode = 0;
/*
 * Since this driver doesn't currently support multicast filtering,
 * set ALL_MULTI if any multicast addresses specified
 */
static int multicast_filter_limit = 0;

static struct usb_eth_dev usb_dev_id[] = {
#define	PEGASUS_DEV(pn, vid, pid, flags)	\
	{name:pn, vendor:vid, device:pid, private:flags},
#include "pegasus.h"
#undef	PEGASUS_DEV
	{NULL, 0, 0, 0}
};

static struct usb_device_id pegasus_ids[] = {
#define	PEGASUS_DEV(pn, vid, pid, flags) \
	{match_flags: USB_DEVICE_ID_MATCH_DEVICE, idVendor:vid, idProduct:pid},
#include "pegasus.h"
#undef	PEGASUS_DEV
	{}
};

MODULE_AUTHOR(DRIVER_AUTHOR);
MODULE_DESCRIPTION(DRIVER_DESC);
MODULE_LICENSE("GPL");
MODULE_PARM(loopback, "i");
MODULE_PARM(mii_mode, "i");
MODULE_PARM_DESC(loopback, "Enable MAC loopback mode (bit 0)");
MODULE_PARM_DESC(mii_mode, "Enable HomePNA mode (bit 0),default=MII mode = 0");

MODULE_DEVICE_TABLE(usb, pegasus_ids);

static int update_eth_regs_async(pegasus_t *);
/* Aargh!!! I _really_ hate such tweaks */
static void ctrl_callback(struct urb *urb)
{
	pegasus_t *pegasus = urb->context;

	if (!pegasus)
		return;

	switch (urb->status) {
	case 0:
		if (pegasus->flags & ETH_REGS_CHANGE) {
			pegasus->flags &= ~ETH_REGS_CHANGE;
			pegasus->flags |= ETH_REGS_CHANGED;
			update_eth_regs_async(pegasus);
			return;
		}
		break;
	case -EINPROGRESS:
		return;
	case -ENOENT:
		break;
	default:
		warn("%s: status %d", __FUNCTION__, urb->status);
	}
	pegasus->flags &= ~ETH_REGS_CHANGED;
	wake_up(&pegasus->ctrl_wait);
}

static int get_registers(pegasus_t * pegasus, __u16 indx, __u16 size,
			 void *data)
{
	int ret;
	unsigned char *buffer;
	DECLARE_WAITQUEUE(wait, current);

	buffer = kmalloc(size, GFP_KERNEL);
	if (!buffer) {
		err("unable to allocate memory for configuration descriptors");
		return 0;
	}
	memcpy(buffer, data, size);

	add_wait_queue(&pegasus->ctrl_wait, &wait);
	set_current_state(TASK_UNINTERRUPTIBLE);
	while (pegasus->flags & ETH_REGS_CHANGED)
		schedule();
	remove_wait_queue(&pegasus->ctrl_wait, &wait);
	set_current_state(TASK_RUNNING);

	pegasus->dr.bRequestType = PEGASUS_REQT_READ;
	pegasus->dr.bRequest = PEGASUS_REQ_GET_REGS;
	pegasus->dr.wValue = cpu_to_le16(0);
	pegasus->dr.wIndex = cpu_to_le16p(&indx);
	pegasus->dr.wLength = cpu_to_le16p(&size);
	pegasus->ctrl_urb->transfer_buffer_length = size;

	FILL_CONTROL_URB(pegasus->ctrl_urb, pegasus->usb,
			 usb_rcvctrlpipe(pegasus->usb, 0),
			 (char *) &pegasus->dr,
			 buffer, size, ctrl_callback, pegasus);

	add_wait_queue(&pegasus->ctrl_wait, &wait);
	set_current_state(TASK_UNINTERRUPTIBLE);

	if ((ret = usb_submit_urb(pegasus->ctrl_urb))) {
		err("%s: BAD CTRLs %d", __FUNCTION__, ret);
		goto out;
	}

	schedule();
out:
	remove_wait_queue(&pegasus->ctrl_wait, &wait);
	memcpy(data, buffer, size);
	kfree(buffer);

	return ret;
}

static int set_registers(pegasus_t * pegasus, __u16 indx, __u16 size,
			 void *data)
{
	int ret;
	unsigned char *buffer;
	DECLARE_WAITQUEUE(wait, current);

	buffer = kmalloc(size, GFP_KERNEL);
	if (!buffer) {
		err("unable to allocate memory for configuration descriptors");
		return 0;
	}
	memcpy(buffer, data, size);

	add_wait_queue(&pegasus->ctrl_wait, &wait);
	set_current_state(TASK_UNINTERRUPTIBLE);
	while (pegasus->flags & ETH_REGS_CHANGED)
		schedule();
	remove_wait_queue(&pegasus->ctrl_wait, &wait);
	set_current_state(TASK_RUNNING);

	pegasus->dr.bRequestType = PEGASUS_REQT_WRITE;
	pegasus->dr.bRequest = PEGASUS_REQ_SET_REGS;
	pegasus->dr.wValue = cpu_to_le16(0);
	pegasus->dr.wIndex = cpu_to_le16p(&indx);
	pegasus->dr.wLength = cpu_to_le16p(&size);
	pegasus->ctrl_urb->transfer_buffer_length = size;

	FILL_CONTROL_URB(pegasus->ctrl_urb, pegasus->usb,
			 usb_sndctrlpipe(pegasus->usb, 0),
			 (char *) &pegasus->dr,
			 buffer, size, ctrl_callback, pegasus);

	add_wait_queue(&pegasus->ctrl_wait, &wait);
	set_current_state(TASK_UNINTERRUPTIBLE);

	if ((ret = usb_submit_urb(pegasus->ctrl_urb))) {
		err("%s: BAD CTRL %d", __FUNCTION__, ret);
		goto out;
	}

	schedule();
out:
	remove_wait_queue(&pegasus->ctrl_wait, &wait);
	kfree(buffer);

	return ret;
}

static int set_register(pegasus_t * pegasus, __u16 indx, __u8 data)
{
	int ret;
	unsigned char *buffer;
	__u16 dat = data;
	DECLARE_WAITQUEUE(wait, current);

	buffer = kmalloc(1, GFP_KERNEL);
	if (!buffer) {
		err("unable to allocate memory for configuration descriptors");
		return 0;
	}
	memcpy(buffer, &data, 1);

	add_wait_queue(&pegasus->ctrl_wait, &wait);
	set_current_state(TASK_UNINTERRUPTIBLE);
	while (pegasus->flags & ETH_REGS_CHANGED)
		schedule();
	remove_wait_queue(&pegasus->ctrl_wait, &wait);
	set_current_state(TASK_RUNNING);

	pegasus->dr.bRequestType = PEGASUS_REQT_WRITE;
	pegasus->dr.bRequest = PEGASUS_REQ_SET_REG;
	pegasus->dr.wValue = cpu_to_le16p(&dat);
	pegasus->dr.wIndex = cpu_to_le16p(&indx);
	pegasus->dr.wLength = cpu_to_le16(1);
	pegasus->ctrl_urb->transfer_buffer_length = 1;

	FILL_CONTROL_URB(pegasus->ctrl_urb, pegasus->usb,
			 usb_sndctrlpipe(pegasus->usb, 0),
			 (char *) &pegasus->dr,
			 buffer, 1, ctrl_callback, pegasus);

	add_wait_queue(&pegasus->ctrl_wait, &wait);
	set_current_state(TASK_UNINTERRUPTIBLE);

	if ((ret = usb_submit_urb(pegasus->ctrl_urb))) {
		err("%s: BAD CTRL %d", __FUNCTION__, ret);
		goto out;
	}

	schedule();
out:
	remove_wait_queue(&pegasus->ctrl_wait, &wait);
	kfree(buffer);

	return ret;
}

static int update_eth_regs_async(pegasus_t * pegasus)
{
	int ret;

	pegasus->dr.bRequestType = PEGASUS_REQT_WRITE;
	pegasus->dr.bRequest = PEGASUS_REQ_SET_REGS;
	pegasus->dr.wValue = 0;
	pegasus->dr.wIndex = cpu_to_le16(EthCtrl0);
	pegasus->dr.wLength = cpu_to_le16(3);
	pegasus->ctrl_urb->transfer_buffer_length = 3;

	FILL_CONTROL_URB(pegasus->ctrl_urb, pegasus->usb,
			 usb_sndctrlpipe(pegasus->usb, 0),
			 (char *) &pegasus->dr,
			 pegasus->eth_regs, 3, ctrl_callback, pegasus);

	if ((ret = usb_submit_urb(pegasus->ctrl_urb)))
		err("%s: BAD CTRL %d, flgs %x", __FUNCTION__, ret,
		    pegasus->flags);

	return ret;
}

static int read_mii_word(pegasus_t * pegasus, __u8 phy, __u8 indx, __u16 * regd)
{
	int i;
	__u8 data[4] = { phy, 0, 0, indx };
	__u16 regdi;

	set_register(pegasus, PhyCtrl, 0);
	set_registers(pegasus, PhyAddr, sizeof(data), data);
	set_register(pegasus, PhyCtrl, (indx | PHY_READ));
	for (i = 0; i < REG_TIMEOUT; i++) {
		get_registers(pegasus, PhyCtrl, 1, data);
		if (data[0] & PHY_DONE)
			break;
	}
	if (i < REG_TIMEOUT) {
		get_registers(pegasus, PhyData, 2, &regdi);
		*regd = le16_to_cpu(regdi);
		return 0;
	}
	warn("%s: failed", __FUNCTION__);

	return 1;
}

static int write_mii_word(pegasus_t * pegasus, __u8 phy, __u8 indx, __u16 regd)
{
	int i;
	__u8 data[4] = { phy, 0, 0, indx };

	data[1] = regd;
	data[2] = regd >> 8;
	set_register(pegasus, PhyCtrl, 0);
	set_registers(pegasus, PhyAddr, 4, data);
	set_register(pegasus, PhyCtrl, (indx | PHY_WRITE));
	for (i = 0; i < REG_TIMEOUT; i++) {
		get_registers(pegasus, PhyCtrl, 1, data);
		if (data[0] & PHY_DONE)
			break;
	}
	if (i < REG_TIMEOUT)
		return 0;
	warn("%s: failed", __FUNCTION__);

	return 1;
}

static int read_eprom_word(pegasus_t * pegasus, __u8 index, __u16 * retdata)
{
	int i;
	__u8 tmp;
	__u16 retdatai;

	set_register(pegasus, EpromCtrl, 0);
	set_register(pegasus, EpromOffset, index);
	set_register(pegasus, EpromCtrl, EPROM_READ);

	for (i = 0; i < REG_TIMEOUT; i++) {
		get_registers(pegasus, EpromCtrl, 1, &tmp);
		if (tmp & EPROM_DONE)
			break;
	}
	if (i < REG_TIMEOUT) {
		get_registers(pegasus, EpromData, 2, &retdatai);
		*retdata = le16_to_cpu(retdatai);
		return 0;
	}
	warn("%s: failed", __FUNCTION__);

	return -1;
}

#ifdef	PEGASUS_WRITE_EEPROM
static inline void enable_eprom_write(pegasus_t * pegasus)
{
	__u8 tmp;

	get_registers(pegasus, EthCtrl2, 1, &tmp);
	set_register(pegasus, EthCtrl2, tmp | EPROM_WR_ENABLE);
}

static inline void disable_eprom_write(pegasus_t * pegasus)
{
	__u8 tmp;

	get_registers(pegasus, EthCtrl2, 1, &tmp);
	set_register(pegasus, EpromCtrl, 0);
	set_register(pegasus, EthCtrl2, tmp & ~EPROM_WR_ENABLE);
}

static int write_eprom_word(pegasus_t * pegasus, __u8 index, __u16 *data)
{
	int	i;
	__u8	tmp;
	__u8 d[4] = { 0x3f, 0, 0, EPROM_WRITE };
	__u16	datai;

	set_registers(pegasus, EpromOffset, 4, d);
	enable_eprom_write(pegasus);
	set_register(pegasus, EpromOffset, index);
	datai = le16_to_cpu(*data);
        d[1] = 0xff & (datai >> 8);
        d[2] = 0xff & datai;
	set_registers( pegasus, EpromData, 2, &d[1] );
	set_register(pegasus, EpromCtrl, EPROM_WRITE);

	for (i = 0; i < REG_TIMEOUT; i++) {
		get_registers(pegasus, EpromCtrl, 1, &tmp);
		if (tmp & EPROM_DONE)
			break;
	}
	disable_eprom_write(pegasus);
	if (i < REG_TIMEOUT)
		return 0;
	warn("%s: failed", __FUNCTION__);
	return -1;
}
#endif				/* PEGASUS_WRITE_EEPROM */

static inline void get_node_id(pegasus_t * pegasus, __u8 * id)
{
	int i;
	__u16 w16;

	for (i = 0; i < 3; i++) {
		read_eprom_word(pegasus, i, &w16);
		((__u16 *) id)[i] = cpu_to_le16p(&w16);
	}
}

static void set_ethernet_addr(pegasus_t * pegasus)
{
	__u8 node_id[6];

	get_node_id(pegasus, node_id);
	set_registers(pegasus, EthID, sizeof(node_id), node_id);
	memcpy(pegasus->net->dev_addr, node_id, sizeof(node_id));
}

static inline int reset_mac(pegasus_t * pegasus)
{
	__u8 data = 0x8;
	int i;

	set_register(pegasus, EthCtrl1, data);
	for (i = 0; i < REG_TIMEOUT; i++) {
		get_registers(pegasus, EthCtrl1, 1, &data);
		if (~data & 0x08) {
			if (loopback & 1)
				break;
			if (mii_mode && (pegasus->features & HAS_HOME_PNA))
				set_register(pegasus, Gpio1, 0x34);
			else
				set_register(pegasus, Gpio1, 0x26);
			set_register(pegasus, Gpio0, pegasus->features);
			set_register(pegasus, Gpio0, DEFAULT_GPIO_SET);
			break;
		}
	}
	if (i == REG_TIMEOUT)
		return 1;

	if (usb_dev_id[pegasus->dev_index].vendor == VENDOR_LINKSYS ||
	    usb_dev_id[pegasus->dev_index].vendor == VENDOR_DLINK) {
		__u16 auxmode;

		read_mii_word(pegasus, 0, 0x1b, &auxmode);
		write_mii_word(pegasus, 0, 0x1b, auxmode | 4);
	}
	if (usb_dev_id[pegasus->dev_index].vendor == VENDOR_ELCON) {
		__u16 auxmode;
		read_mii_word(pegasus, 3, 0x1b, &auxmode);
		write_mii_word(pegasus, 3, 0x1b, auxmode | 4);
	}
	return 0;
}


static void enable_net_traffic( pegasus_t *pegasus )
{
	__u16	linkpart, bmsr, bmcr;
	__u8 data[4];

	if ( read_mii_word(pegasus, pegasus->phy, MII_BMCR, &bmcr) ) 
		return;
	if ( read_mii_word(pegasus, pegasus->phy, MII_BMSR, &bmsr) ) 
		return;

	if ( !(bmsr & BMSR_ANEGCOMPLETE) && !loopback ) 
		warn( "%s: link NOT established (0x%x) - check the cable.",
			pegasus->net->name, bmsr );

	if ( (bmcr & BMCR_ANENABLE) &&
	     (bmsr & (BMSR_ANEGCOMPLETE|BMSR_ANEGCAPABLE|BMSR_LSTATUS)) ) {
	if (read_mii_word(pegasus, pegasus->phy, MII_LPA, &linkpart))
		    return;
	if (!(linkpart & 1))
		warn("link partner stat %x", linkpart);
	}
	else
		linkpart = 0;

	data[0] = 0xc9; // txen, rxen, rxstat-append, crc-in-rxpkt
	if (pegasus->eth_regs[EthCtrl0] | RX_MULTICAST)
		data[0] |= RX_MULTICAST;
	data[1] = 0;
	if ( linkpart & (LPA_100FULL | LPA_10FULL) )
		data[1] |= 0x20;	/* set full duplex */
	if ( linkpart & (LPA_100FULL | LPA_100HALF) )
		data[1] |= 0x10;	/* set 100 Mbps */
	if (mii_mode)
		data[1] = 0;
	data[2] = (loopback & 1) ? 
		0x09 :          // loop_back, ep3-rd-clr-2b_2f&7a 
		0x01;           // ep3-rd-clr-2b_2f&7a

	memcpy(pegasus->eth_regs, data, sizeof(data));

	set_registers(pegasus, EthCtrl0, 3, data);
	if (pegasus->link_sts && LINKSYS_USB100TX_LED_WAR(pegasus)) {
		LINKSYS_ENA_TXRXLED;
		if (data[1] & 0x10) {
			LINKSYS_GREEN_LINKLED;
		}
		else {
			LINKSYS_AMBER_LINKLED;
		}
	}
	info("link speed and mode 10%s%sDX",
			(data[1]&0x10)?"0":"", (data[1]&0x20)?"F":"H");

}

static void read_bulk_callback(struct urb *urb)
{
	pegasus_t *pegasus = urb->context;
	struct net_device *net;
	int count = urb->actual_length, res;
	__u8 *rx_buffp, rx_stat;
	__u16 rx_len;
	struct sk_buff *skb;
	__u16 pkt_len;

	if (!pegasus || !(pegasus->flags & PEGASUS_RUNNING)) {
		warn("!dev");
		return;
	}

	net = pegasus->net;
	if (!netif_device_present(net)) {
		warn("netdev is not present");
		return;
	}

	if (pegasus->flags & PEGASUS_RX_BUSY) {
		pegasus->stats.rx_errors++;
		warn("pegasus Rx busy");
		return;
	}
	pegasus->flags |= PEGASUS_RX_BUSY;

	switch (urb->status) {
	case 0:
		break;
	case -ETIMEDOUT:
		warn("RX timeout, reset MAC");
		pegasus->flags &= ~PEGASUS_RX_BUSY;
		return;
	default:
		warn( "%s: RX status %d", net->name, urb->status );
		goto goon;
	}

	if (!count)
		goto goon;

	rx_buffp = pegasus->rx_buff + count - 4;
	rx_len = rx_buffp[0] | rx_buffp[1] << 8;
	rx_stat = rx_buffp[2];
	if (rx_stat & 0x0e) {
		dbg("%s: RX packet error %x", net->name, rx_stat & 0x0e);
		pegasus->stats.rx_errors++;
		if (rx_stat & 0x06)
			pegasus->stats.rx_length_errors++;
		if (rx_stat & 0x08)
			pegasus->stats.rx_crc_errors++;
		if (rx_stat & 0x10)
			pegasus->stats.rx_frame_errors++;
		goto goon;
	}

	pkt_len = (rx_len & 0xfff) - 8;

	if (!(skb = dev_alloc_skb(pkt_len + 2)))
		goto goon;

	skb->dev = net;
	skb_reserve(skb, 2);
	eth_copy_and_sum(skb, pegasus->rx_buff, pkt_len, 0);
	skb_put(skb, pkt_len);

	skb->protocol = eth_type_trans(skb, net);
	netif_rx(skb);
	pegasus->stats.rx_packets++;
	pegasus->stats.rx_bytes += pkt_len;

goon:
	FILL_BULK_URB(pegasus->rx_urb, pegasus->usb,
		      usb_rcvbulkpipe(pegasus->usb, 1),
		      pegasus->rx_buff, PEGASUS_MAX_MTU,
		      read_bulk_callback, pegasus);
	if ((res = usb_submit_urb(pegasus->rx_urb)))
		warn("%s: failed submint rx_urb %d", __FUNCTION__, res);
	pegasus->flags &= ~PEGASUS_RX_BUSY;
}

static void write_bulk_callback(struct urb *urb)
{
	pegasus_t *pegasus = urb->context;

	if (!pegasus || !(pegasus->flags & PEGASUS_RUNNING))
		return;

	if (!netif_device_present(pegasus->net))
		return;

	if (urb->status)
		info("%s: TX status %d", pegasus->net->name, urb->status);

	pegasus->net->trans_start = jiffies;
	netif_wake_queue(pegasus->net);
}

#ifdef	PEGASUS_USE_INTR
static void intr_callback(struct urb *urb)
{
	pegasus_t *pegasus = urb->context;
	struct net_device *net;
	__u8 *d;

	if (!pegasus)
		return;

	switch (urb->status) {
	case 0:
		break;
	case -ENOENT:
#ifdef DEBUG
		if (msglvl > 8)
			dbg("USB_ST_URB_KILLED");
#endif
		return;
	default:
#ifdef DEBUG
		if (msglvl > 8)
			dbg("intr status %d", urb->status);
#endif
#if 0 // too verbose...
		info("intr status %d", urb->status);
#endif
		break;
	}

	if (pegasus->flags & PEGASUS_UNPLUG) {
		warn("UNPLUG intr, sched pending %d", pegasus->sched_pending);
	}

	d = urb->transfer_buffer;
	net = pegasus->net;
	if (d[0] & 0xfc) {
		pegasus->stats.tx_errors++;
		if (d[0] & TX_UNDERRUN)
			pegasus->stats.tx_fifo_errors++;
		if ( d[0] & EXCESSIVE_COL )
			pegasus->stats.collisions++;
		if ( d[0] & (JABBER_TIMEOUT) )
			pegasus->stats.tx_aborted_errors++;
		if (d[0] & LATE_COL)
			pegasus->stats.tx_window_errors++;
		if (d[0] & (NO_CARRIER | LOSS_CARRIER))
			pegasus->stats.tx_carrier_errors++;
	}
	if ( d[1] & TX_FIFOFULL )
		pegasus->stats.tx_dropped++;

	if ( d[2] & RX_OVERFLOW ) {
		unsigned long dropped;

		pegasus->stats.rx_errors++;
		pegasus->stats.rx_fifo_errors++;
		dropped = ((d[3] << 8) | d[4]);
		pegasus->stats.rx_dropped += dropped;
	}
	if ( !(d[5] & WAKEUP_LINK_STS) ) {
		if (pegasus->link_sts) {
			pegasus->link_down++;
			pegasus->link_sts = 0;
			netif_carrier_off(pegasus->net);
			spin_lock(&pegasus->intr_lock);
			pegasus->sched_pending++;
			spin_unlock(&pegasus->intr_lock);
			schedule_task(&pegasus->sched_tq_dis);
			info("link down");
		}
	}
	else {
		if (!pegasus->link_sts) {
			pegasus->link_up++;
			pegasus->link_sts = 1;
			netif_carrier_on(pegasus->net);
			spin_lock(&pegasus->intr_lock);
			pegasus->sched_pending++;
			spin_unlock(&pegasus->intr_lock);
			schedule_task(&pegasus->sched_tq_ena);
			info("link up");
		}
	}
	pegasus->poll_cnt++;
}
#endif

static void pegasus_tx_timeout(struct net_device *net)
{
	pegasus_t *pegasus = net->priv;

	if (!pegasus)
		return;

	warn("%s: Tx timed out.", net->name);
	pegasus->tx_urb->transfer_flags |= USB_ASYNC_UNLINK;
	usb_unlink_urb(pegasus->tx_urb);
	pegasus->stats.tx_errors++;
}

static int pegasus_start_xmit(struct sk_buff *skb, struct net_device *net)
{
	pegasus_t *pegasus = net->priv;
	int 	count = skb->len + 2;
	int res;
	__u16 l16 = skb->len;

	netif_stop_queue(net);

	if (count & 0x3f) {
		pegasus->tx_urb->transfer_flags &= ~USB_ZERO_PACKET;
	} else {
		pegasus->tx_urb->transfer_flags |= USB_ZERO_PACKET;
	}

	((__u16 *) pegasus->tx_buff)[0] = cpu_to_le16(l16);
	memcpy(pegasus->tx_buff + 2, skb->data, skb->len);
	FILL_BULK_URB(pegasus->tx_urb, pegasus->usb,
		      usb_sndbulkpipe(pegasus->usb, 2),
		      pegasus->tx_buff, PEGASUS_MAX_MTU,
		      write_bulk_callback, pegasus);
	pegasus->tx_urb->transfer_buffer_length = count;
	if ((res = usb_submit_urb(pegasus->tx_urb))) {
		warn("failed tx_urb %d", res);
		pegasus->stats.tx_errors++;
		netif_start_queue(net);
	} else {
		pegasus->stats.tx_packets++;
		pegasus->stats.tx_bytes += skb->len;
		net->trans_start = jiffies;
	}

	dev_kfree_skb(skb);

	return 0;
}

static struct net_device_stats *pegasus_netdev_stats(struct net_device *dev)
{
#if defined(PEGASUS_USE_INTR) && defined(DEBUG)
	pegasus_t *p = (pegasus_t *) dev->priv;
	dbg("link=%d up=%lu dn=%lu poll_cnt=%lu",
			p->link_sts, p->link_up, p->link_down, p->poll_cnt);
#endif
	return &((pegasus_t *) dev->priv)->stats;
}

static inline void disable_net_traffic(pegasus_t * pegasus)
{
	int tmp = 0;

	set_registers(pegasus, EthCtrl0, 2, &tmp);
	if (LINKSYS_USB100TX_LED_WAR(pegasus)) {
		LINKSYS_DIS_TXRXLED;
		LINKSYS_DIS_LINKLED;
	}
}


#ifdef PEGASUS_USE_INTR
static void enable_net_traffic_sched( void *arg )
{
	pegasus_t *pegasus = arg;
	unsigned long flags;
	unsigned int pending;

	down(&pegasus->sem);
	spin_lock_irqsave(&pegasus->intr_lock, flags);
	--pegasus->sched_pending;
	if (pegasus->flags & PEGASUS_UNPLUG) {
		pending = pegasus->sched_pending;
		spin_unlock_irqrestore(&pegasus->intr_lock, flags);
		up(&pegasus->sem);
		if (pending == 0)
			kfree(pegasus);
		return;
	}
	spin_unlock_irqrestore(&pegasus->intr_lock, flags);

	enable_net_traffic(pegasus);
	up(&pegasus->sem);
}

static void disable_net_traffic_sched( void *arg )
{
	pegasus_t *pegasus = arg;
	unsigned long flags;
	unsigned int pending;

	down(&pegasus->sem);
	spin_lock_irqsave(&pegasus->intr_lock, flags);
	--pegasus->sched_pending;
	if (pegasus->flags & PEGASUS_UNPLUG) {
		pending = pegasus->sched_pending;
		spin_unlock_irqrestore(&pegasus->intr_lock, flags);
		up(&pegasus->sem);
		if (pending == 0)
			kfree(pegasus);
		return;
	}
	spin_unlock_irqrestore(&pegasus->intr_lock, flags);

	disable_net_traffic(pegasus);
	up(&pegasus->sem);
}
#endif

static inline void get_interrupt_interval(pegasus_t * pegasus)
{
	__u8 data[2];

	read_eprom_word(pegasus, 4, (__u16 *) data);
	if (data[1] < 0x80) {
		info("intr interval will be changed from %ums to %ums",
		     data[1], 0x80);
		data[1] = 0x80;
#ifdef	PEGASUS_WRITE_EEPROM
		write_eprom_word(pegasus, 4, (__u16 *) data);
#endif
	}
	pegasus->intr_interval = data[1];
}

#if 0
static void set_carrier(struct net_device *net)
{
	pegasus_t *pegasus;
	short tmp;

	pegasus = net->priv;
	read_mii_word(pegasus, pegasus->phy, MII_BMSR, &tmp);
	if (tmp & BMSR_LSTATUS)
		netif_carrier_on(net);
	else
		netif_carrier_off(net);

}
#endif

static int pegasus_open(struct net_device *net)
{
	pegasus_t *pegasus = (pegasus_t *) net->priv;
	int res;

	dbg("<<");
	down(&pegasus->sem);
	FILL_BULK_URB(pegasus->rx_urb, pegasus->usb,
		      usb_rcvbulkpipe(pegasus->usb, 1),
		      pegasus->rx_buff, PEGASUS_MAX_MTU,
		      read_bulk_callback, pegasus);
	if ((res = usb_submit_urb(pegasus->rx_urb)))
		warn("%s: failed rx_urb %d", __FUNCTION__, res);
#ifdef	PEGASUS_USE_INTR
	FILL_INT_URB(pegasus->intr_urb, pegasus->usb,
		     usb_rcvintpipe(pegasus->usb, 3),
		     pegasus->intr_buff, sizeof(pegasus->intr_buff),
		     intr_callback, pegasus, pegasus->intr_interval);
	if ((res = usb_submit_urb(pegasus->intr_urb)))
		warn("%s: failed intr_urb %d", __FUNCTION__, res);
#endif
	netif_start_queue(net);
	enable_net_traffic(pegasus);
	pegasus->flags |= PEGASUS_RUNNING;
	up(&pegasus->sem);

	dbg(">>");
	return 0;
}

static int pegasus_close(struct net_device *net)
{
	pegasus_t *pegasus = net->priv;

	dbg("<<");
	if ( !pegasus) {
		dbg("no device handle");
		return -ENODEV;
	}

	down(&pegasus->sem);
	disable_net_traffic(pegasus);
	pegasus->flags &= ~PEGASUS_RUNNING;
	netif_stop_queue(net);

	usb_unlink_urb(pegasus->rx_urb);
	usb_unlink_urb(pegasus->tx_urb);
	usb_unlink_urb(pegasus->ctrl_urb);
#ifdef	PEGASUS_USE_INTR
	usb_unlink_urb(pegasus->intr_urb);
#endif
	up(&pegasus->sem);
	dbg(">>");
	return 0;
}

#define FIRST_REG		0
#define LAST_REG		(0x81)
#define MII_REGS_SIZE		((MII_EXPANSION - MII_BMCR + 1)*2)
#define PEGASUS_REGS_SIZE	(LAST_REG - FIRST_REG + MII_REGS_SIZE + 1)
#define PEGASUS_REGS_VER	1
#define PEGASUS_EEPROM_SIZE	(0x5C)

static int pegasus_ethtool_ioctl(struct net_device *dev, void *uaddr)
{
	pegasus_t *pegasus;
	int cmd;
	char tmp[128];

	pegasus = dev->priv;
	if (get_user(cmd, (int *) uaddr))
		return -EFAULT;

	dbg( "cmd = %x", cmd); 
	switch (cmd) {
	case ETHTOOL_GDRVINFO:{
		struct ethtool_drvinfo info = { ETHTOOL_GDRVINFO };
		strncpy(info.driver, DRIVER_DESC, ETHTOOL_BUSINFO_LEN);
		strncpy(info.version, DRIVER_VERSION,
			ETHTOOL_BUSINFO_LEN);
		sprintf(tmp, "usb%d:%d", pegasus->usb->bus->busnum,
			pegasus->usb->devnum);
		strncpy(info.bus_info, tmp, ETHTOOL_BUSINFO_LEN);
		info.eedump_len = PEGASUS_EEPROM_SIZE;
		info.regdump_len = PEGASUS_REGS_SIZE;
		if (copy_to_user(uaddr, &info, sizeof(info)))
			return -EFAULT;
		return 0;
		}
	case ETHTOOL_GSET:{
		struct ethtool_cmd ecmd;
		short lpa, bmcr, ana;

		if (copy_from_user(&ecmd, uaddr, sizeof(ecmd)))
			return -EFAULT;
		ecmd.supported = (SUPPORTED_10baseT_Half |
				  SUPPORTED_10baseT_Full |
				  SUPPORTED_100baseT_Half |
				  SUPPORTED_100baseT_Full |
				  SUPPORTED_Autoneg |
				  SUPPORTED_TP | SUPPORTED_MII);
		ecmd.port = PORT_TP;
		ecmd.transceiver = XCVR_INTERNAL;
		ecmd.phy_address = pegasus->phy;
		read_mii_word(pegasus, pegasus->phy, MII_BMCR, &bmcr);
		if (bmcr & BMCR_ANENABLE) {
			read_mii_word(pegasus, pegasus->phy, MII_LPA, &lpa);
			read_mii_word(pegasus, pegasus->phy, MII_ADVERTISE, &ana);
			ecmd.advertising = ADVERTISED_Autoneg;
			ecmd.advertising |= (ana | ADVERTISE_10HALF) ? ADVERTISED_10baseT_Half : 0;
			ecmd.advertising |= (ana | ADVERTISE_10FULL) ? ADVERTISED_10baseT_Full : 0;
			ecmd.advertising |= (ana | ADVERTISE_100HALF) ? ADVERTISED_100baseT_Half : 0;
			ecmd.advertising |= (ana | ADVERTISE_100FULL) ? ADVERTISED_100baseT_Full : 0;
			ecmd.autoneg = AUTONEG_ENABLE;
			ecmd.speed = lpa & (LPA_100HALF | LPA_100FULL) ?
				    SPEED_100 : SPEED_10;
			if (ecmd.speed == SPEED_100)
				ecmd.duplex = lpa & LPA_100FULL ?
					    DUPLEX_FULL : DUPLEX_HALF;
			else
				ecmd.duplex = lpa & LPA_10FULL ?
					    DUPLEX_FULL : DUPLEX_HALF;
		} else {
			ecmd.autoneg = AUTONEG_DISABLE;
			ecmd.speed = bmcr & BMCR_SPEED100 ?
				SPEED_100 : SPEED_10;
			ecmd.duplex = bmcr & BMCR_FULLDPLX ?
				DUPLEX_FULL : DUPLEX_HALF;
		}
		if (copy_to_user(uaddr, &ecmd, sizeof(ecmd)))
			return -EFAULT;

		return 0;
		}
	case ETHTOOL_SSET:{
		return -EOPNOTSUPP;
		}
	case ETHTOOL_GLINK:{
		struct ethtool_value edata = { ETHTOOL_GLINK };
		edata.data = netif_carrier_ok(dev);
		if (copy_to_user(uaddr, &edata, sizeof(edata)))
			return -EFAULT;
		return 0;
		}
	/* get registers */
	case ETHTOOL_GREGS: {
		struct ethtool_regs regs;
		u8 regbuf[PEGASUS_REGS_SIZE];
		u8 d[4];
		int r, i, x;

		if (copy_from_user(&regs, uaddr, sizeof(regs)))
			return -EFAULT;

		if (regs.len > PEGASUS_REGS_SIZE) {
			regs.len = PEGASUS_REGS_SIZE;
		}

		regs.version = PEGASUS_REGS_VER;
		if (copy_to_user(uaddr, &regs, sizeof(regs)))
			return -EFAULT;


		x = FIRST_REG;
		i = 0;
		while ( x < LAST_REG  ) {
			r = get_registers(pegasus, x, 2, d);
			dbg("i=%d x=%x d0=%x d1=%x r=%d", 
				i, x, d[0], d[1], r);
			regbuf[i] = d[0];
			regbuf[i+1] = d[1];
			i += 2;
			x += 2;
		}
		x = 0;
		while ( x <= 6  ) {
			read_mii_word(pegasus, pegasus->phy, x, (__u16 *)d);
			regbuf[i] = d[0];
			regbuf[i+1] = d[1];
			dbg("mii  x=%x %02x%02x", 
				x, d[0], d[1]);
			i += 2;
			x += 1;
		}

		uaddr += offsetof(struct ethtool_regs, data);
		if (copy_to_user(uaddr, regbuf, regs.len))
			return -EFAULT;
		return 0;
	}

#ifdef DEBUG
	/* get eprom */
	case ETHTOOL_GEEPROM: {
		struct ethtool_eeprom eeprom;
		u8 eprombuf[PEGASUS_EEPROM_SIZE];
		u8 d[4];
		int r, i, x;

		if (copy_from_user(&eeprom, uaddr, sizeof(eeprom)))
			return -EFAULT;

		if (eeprom.offset + eeprom.len > PEGASUS_EEPROM_SIZE) {
			return -EFAULT;
		}

		x = eeprom.offset;
		i = 0;
		while ( x < PEGASUS_EEPROM_SIZE/2  ) {
			r = read_eprom_word(pegasus, x, (__u16 *)d );
			dbg("eprom x=%x d0=%x d1=%x r=%d", 
				x, d[0], d[1], r);
			eprombuf[i] = d[0];
			eprombuf[i+1] = d[1];
			i += 2;
			x += 1;
		}
		uaddr += offsetof(struct ethtool_eeprom, data);
		if (copy_to_user(uaddr, eprombuf, eeprom.len))
			return -EFAULT;
		return 0;
	}

    	/* get message-level */
	case ETHTOOL_GMSGLVL: {
		struct ethtool_value edata = {ETHTOOL_GMSGLVL};
		edata.data = msglvl;
		dbg("link=%d up=%lu dn=%lu poll_cnt=%lu pend=%d",
		    pegasus->link_sts, pegasus->link_up, pegasus->link_down, pegasus->poll_cnt,
		    pegasus->sched_pending);
		if (copy_to_user(uaddr, &edata, sizeof(edata)))
			return -EFAULT;
		return 0;
	}
#endif
	/* set message-level */
	case ETHTOOL_SMSGLVL: {
		struct ethtool_value edata;
		void usb_hub_port_reconnect(struct usb_device *);

		if (copy_from_user(&edata, uaddr, sizeof(edata)))
			return -EFAULT;
		if (edata.data == 99)
			usb_hub_port_reconnect(pegasus->usb);
#ifdef DEBUG
		else
			msglvl = edata.data; 

		if (edata.data > 0x70FF && edata.data < 0x7fff)
		{
			__u8  val = edata.data & 0xff;
			__u16 idx = (edata.data >> 8) & 0xffff;

			set_register(pegasus, idx, val);
			printk("set register [0x%x] = 0x%x\n", idx, val);
		}

		if (edata.data == 199) {
			static int gpio;
			__u8 val;
			int r7c, r7e, r7f;

			val = 0x24;
			if (gpio & 0x01) val |= 2;
			if (gpio & 0x02) val |= 0x10;
			set_register(pegasus, 0x7c, val);
			val = 0x24;
			if (gpio & 0x04) val |= 2;
			if (gpio & 0x08) val |= 0x10;
			set_register(pegasus, 0x7e, val);
			val = 0x34;
			if (gpio & 0x10) val |= 2;
			if (gpio & 0x20) val &= ~0x10;
			set_register(pegasus, 0x7f, val);
			r7c = r7e = r7f = 0;
			get_registers(pegasus, 0x7c, 1, &r7c);
			get_registers(pegasus, 0x7e, 1, &r7e);
			get_registers(pegasus, 0x7f, 1, &r7f);
			printk("gpio#=0x%x [7c]=%x [7e]=%x [7f]=%x\n",
					gpio, r7c, r7e, r7f);

			gpio++;
		}
#endif
		return 0;
	}
	/* restart autonegotiation */
	case ETHTOOL_NWAY_RST: {
		__u16 bmcr;
		info("restarting auto-negotiations");
		if ( read_mii_word(pegasus, pegasus->phy, MII_BMCR, &bmcr) ) 
			return -EFAULT;
		bmcr |= BMCR_ANRESTART;
		if ( write_mii_word(pegasus, pegasus->phy, MII_BMCR, bmcr) ) 
			return -EFAULT;
		return 0;
	}
	default:
		return -EOPNOTSUPP;
	}
	return 0;
}

static int pegasus_ioctl(struct net_device *net, struct ifreq *rq, int cmd)
{
	__u16 *data = (__u16 *) & rq->ifr_data;
	pegasus_t *pegasus = net->priv;
	int res;

	res = 0;

	down(&pegasus->sem);
	switch (cmd) {
	case SIOCETHTOOL:
		res = pegasus_ethtool_ioctl(net, rq->ifr_data);
		break;
	case SIOCDEVPRIVATE:
		data[0] = pegasus->phy;
	case SIOCDEVPRIVATE+1:
		read_mii_word(pegasus, data[0], data[1] & 0x1f, &data[3]);
		break;
	case SIOCDEVPRIVATE + 2:
		if (!capable(CAP_NET_ADMIN)) {
			res = -EPERM;
			break;
		}
		write_mii_word(pegasus, pegasus->phy, data[1] & 0x1f, data[2]);
		break;
	default:
		res = -EOPNOTSUPP;
	}
	up(&pegasus->sem);

	return res;
}

static void pegasus_set_multicast(struct net_device *net)
{
	pegasus_t *pegasus = net->priv;

	if (net->flags & IFF_PROMISC) {
		pegasus->eth_regs[EthCtrl2] |= RX_PROMISCUOUS;
		info("%s: Promiscuous mode enabled", net->name);
	} else if ((net->mc_count > multicast_filter_limit) ||
		   (net->flags & IFF_ALLMULTI)) {
		pegasus->eth_regs[EthCtrl0] |= RX_MULTICAST;
		pegasus->eth_regs[EthCtrl2] &= ~RX_PROMISCUOUS;
		info("%s set allmulti", net->name);
	} else {
		pegasus->eth_regs[EthCtrl0] &= ~RX_MULTICAST;
		pegasus->eth_regs[EthCtrl2] &= ~RX_PROMISCUOUS;
	}

	pegasus->flags |= ETH_REGS_CHANGE;
	ctrl_callback(pegasus->ctrl_urb);
}

static __u8 mii_phy_probe(pegasus_t * pegasus)
{
	int i;
	__u16 tmp;

	for (i = 0; i < 32; i++) {
		read_mii_word(pegasus, i, MII_BMSR, &tmp);
		if (tmp == 0 || tmp == 0xffff || (tmp & BMSR_MEDIA) == 0)
			continue;
		else
			return i;
	}

	return 0xff;
}

static inline void setup_pegasus_II(pegasus_t * pegasus)
{
	set_register(pegasus, Reg1d, 0);
	set_register(pegasus, Reg7b, 2);
	if (pegasus->features & HAS_HOME_PNA && mii_mode)
		set_register(pegasus, Reg81, 6);
	else
		set_register(pegasus, Reg81, 2);
}

static void *pegasus_probe(struct usb_device *dev, unsigned int ifnum,
			   const struct usb_device_id *id)
{
	struct net_device *net;
	pegasus_t *pegasus;
	int dev_index = id - pegasus_ids;

	dbg("<<[%d] r=%d", dev->devnum, dev->refcnt.counter);
	if (usb_set_configuration(dev, dev->config[0].bConfigurationValue)) {
		err("usb_set_configuration() failed");
		pegasus = NULL;
		goto exit;
	}

	if (!(pegasus = kmalloc(sizeof(struct pegasus), GFP_KERNEL))) {
		err("out of memory allocating device structure");
		goto exit;
	}

	usb_inc_dev_use(dev);
	memset(pegasus, 0, sizeof(struct pegasus));
	pegasus->dev_index = dev_index;
	init_waitqueue_head(&pegasus->ctrl_wait);

	pegasus->ctrl_urb = usb_alloc_urb(0);
	if (!pegasus->ctrl_urb) {
		kfree(pegasus);
		pegasus = NULL;
		goto exit;
	}
	pegasus->rx_urb = usb_alloc_urb(0);
	if (!pegasus->rx_urb) {
		usb_free_urb(pegasus->ctrl_urb);
		kfree(pegasus);
		pegasus = NULL;
		goto exit;
	}
	pegasus->tx_urb = usb_alloc_urb(0);
	if (!pegasus->tx_urb) {
		usb_free_urb(pegasus->rx_urb);
		usb_free_urb(pegasus->ctrl_urb);
		kfree(pegasus);
		pegasus = NULL;
		goto exit;
	}
#ifdef	PEGASUS_USE_INTR
	pegasus->intr_urb = usb_alloc_urb(0);
	if (!pegasus->intr_urb) {
		usb_free_urb(pegasus->tx_urb);
		usb_free_urb(pegasus->rx_urb);
		usb_free_urb(pegasus->ctrl_urb);
		kfree(pegasus);
		pegasus = NULL;
		goto exit;
	}
#endif

	net = init_etherdev(NULL, 0);
	if (!net) {
		usb_free_urb(pegasus->tx_urb);
		usb_free_urb(pegasus->rx_urb);
		usb_free_urb(pegasus->ctrl_urb);
#ifdef	PEGASUS_USE_INTR
		usb_free_urb(pegasus->intr_urb);
#endif
		kfree(pegasus);
		err("out of memory allocating etherdev structure");
		usb_dec_dev_use(dev);
		pegasus = NULL;
		goto exit;
	}

	spin_lock_init(&pegasus->intr_lock);
	init_MUTEX(&pegasus->sem);
	pegasus->usb = dev;
	pegasus->net = net;
	SET_MODULE_OWNER(net);
	net->priv = pegasus;
	net->open = pegasus_open;
	net->stop = pegasus_close;
	net->watchdog_timeo = PEGASUS_TX_TIMEOUT;
	net->tx_timeout = pegasus_tx_timeout;
	net->do_ioctl = pegasus_ioctl;
	net->hard_start_xmit = pegasus_start_xmit;
	net->set_multicast_list = pegasus_set_multicast;
	net->get_stats = pegasus_netdev_stats;
	net->mtu = PEGASUS_MTU;

	pegasus->features = usb_dev_id[dev_index].private;
#ifdef	PEGASUS_USE_INTR
	get_interrupt_interval(pegasus);
	pegasus->sched_tq_dis.routine = disable_net_traffic_sched;
	pegasus->sched_tq_dis.data = (void *)pegasus;
	pegasus->sched_tq_ena.routine = enable_net_traffic_sched;
	pegasus->sched_tq_ena.data = (void *)pegasus;
#endif
	if (reset_mac(pegasus)) {
		err("can't reset MAC");
		unregister_netdev(pegasus->net);
		usb_free_urb(pegasus->tx_urb);
		usb_free_urb(pegasus->rx_urb);
		usb_free_urb(pegasus->ctrl_urb);
		kfree(pegasus->net);
		kfree(pegasus);
		usb_dec_dev_use(dev);
		pegasus = NULL;
		goto exit;
	}

	info("%s: %s", net->name, usb_dev_id[dev_index].name);

	netif_carrier_off(pegasus->net);
	set_ethernet_addr(pegasus);

	if (pegasus->features & PEGASUS_II) {
		info("setup Pegasus II specific registers");
		setup_pegasus_II(pegasus);
	}

	pegasus->phy = mii_phy_probe(pegasus);
	if (pegasus->phy == 0xff) {
		warn("can't locate MII phy, using default");
		pegasus->phy = 1;
	}
	dbg( " MII phy=%d usb_dev=%p", pegasus->phy, pegasus->usb);

exit:
	dbg(">>[%d] r=%d p=%p", dev->devnum, dev->refcnt.counter, pegasus);
	return pegasus;
}

static void pegasus_disconnect(struct usb_device *dev, void *ptr)
{
	struct pegasus *pegasus = ptr;
	unsigned long flags;
	unsigned int pending;

	dbg("<<[%d] r=%d ptr=%p", dev->devnum, dev->refcnt.counter, pegasus);
	if (!pegasus) {
		warn("disconnecting non-existant device");
		return;
	}

	spin_lock_irqsave(&pegasus->intr_lock, flags);
	pegasus->flags |= PEGASUS_UNPLUG;
	spin_unlock_irqrestore(&pegasus->intr_lock, flags);
	unregister_netdev(pegasus->net);
	usb_free_urb(pegasus->tx_urb);
	usb_free_urb(pegasus->rx_urb);
	usb_free_urb(pegasus->ctrl_urb);
#ifdef	PEGASUS_USE_INTR
	usb_free_urb(pegasus->intr_urb);
#endif
	usb_dec_dev_use(dev);
	spin_lock_irqsave(&pegasus->intr_lock, flags);
	pending = pegasus->sched_pending;
	spin_unlock_irqrestore(&pegasus->intr_lock, flags);

	pegasus->net->priv = NULL;
	kfree(pegasus->net);
	pegasus->net = NULL;

	if (pending == 0)
		kfree(pegasus);
	dbg(">>[%d] r=%d pend=%d", dev->devnum, dev->refcnt.counter, pending);
}

static struct usb_driver pegasus_driver = {
	name:		"pegasus",
	probe:		pegasus_probe,
	disconnect:	pegasus_disconnect,
	id_table:	pegasus_ids,
};

int __init pegasus_init(void)
{
	int res;
	dbg("<<");
	info(DRIVER_VERSION ":" DRIVER_DESC);
	res = usb_register(&pegasus_driver);
	dbg(">>");
	return res;
}

void __exit pegasus_exit(void)
{
	dbg("<<");
	usb_deregister(&pegasus_driver);
	dbg(">>");
}

module_init(pegasus_init);
module_exit(pegasus_exit);
